/*
 * Copyright 2024 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.tig.client;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import dk.vajhoej.vms.tig.Arg;
import dk.vajhoej.vms.tig.Arg.DataType;
import dk.vajhoej.vms.tig.Arg.Usage;
import dk.vajhoej.vms.tig.Method;

public class CClientGen implements ClientGen {
	private PrintWriter pw_c;
	private PrintWriter pw_h;
	@Override
	public String fileName(String imgnam) {
		return className(imgnam) + ".c";
	}
	@Override
	public String className(String imgnam) {
		return imgnam.toLowerCase() + "_client";
	}
	@Override
	public String typeName(Arg.DataType typ) {
		switch(typ) {
			case Byte:
				return "int8_t";
			case Word:
				return "int16_t";
			case LongWord:
				return "int32_t";
			case QuadWord:
				return "int64_t";
			case F_Float:
			case S_Float:
				return "float";
			case D_Float:
			case G_Float:
			case T_Float:
				return "double";
			case FixedCharacterString:
			case VariableCharacterString:
			case NullTermCharacterString:
				return "charptr";
			case Block:
				return "block";
			default:
				throw new RuntimeException("Unsupported type: " + typ);
		}
	}
	@Override
	public void open(String fnm) throws IOException {
		pw_c = new PrintWriter(new FileWriter(fnm));
		String fnm2 = fnm.substring(0, fnm.length() - 2) + ".h";
		pw_h = new PrintWriter(new FileWriter(fnm2));
	}
	private String prefix;
	@Override
	public void genClassHeader(String ns, String clznam) {
		prefix = clznam;
		pw_h.printf("// GENERATED CODE - edits may be overwritten\n");
		pw_h.printf("#ifndef %s_H\n", prefix.toUpperCase());
		pw_h.printf("#define %s_H\n", prefix.toUpperCase());
		pw_h.printf("\n");
		pw_h.printf("struct tig_t\n");
		pw_h.printf("{\n");
		pw_h.printf("    const char* host;\n");
		pw_h.printf("    int port;\n");
		pw_h.printf("    struct con_t *con;\n");
		pw_h.printf("};\n");
		pw_h.printf("\n");
		pw_h.printf("typedef const char *charptr;\n");
		pw_h.printf("\n");
		pw_c.printf("// GENERATED CODE - edits may be overwritten\n");
		pw_c.printf("#include <stdint.h>\n");
		pw_c.printf("\n");
		pw_c.printf("#include \"tig.h\"\n");
		pw_c.printf("\n");
		pw_c.printf("#include \"%s.h\"\n", clznam);
		pw_c.printf("\n");
	}
	@Override
	public void genCtor(String clznam, int port) {
		pw_c.printf("void %s_init(struct tig_t *ctx, const char* host)\n", prefix);
		pw_c.printf("{\n");
		pw_c.printf("    ctx->host = host;\n");
		pw_c.printf("    ctx->port = %d;\n", port);
		pw_c.printf("}\n");
		pw_c.printf("\n");
		pw_h.printf("void %s_init(struct tig_t *ctx, const char* host);\n", prefix);
	}
	@Override
	public void genMethodHeader(String methnam, Method.ArgList args) {
		pw_c.printf("int32_t %s_%s(struct tig_t *ctx", prefix, methnam.toLowerCase());
		for(Arg arg : args.getArgs()) {
			pw_c.printf(", %s", arg.getType() != Arg.DataType.Block ? typeName(arg.getType()) : arg.getImpl().toLowerCase());
			if(arg.getUse() != Usage.In) {
				// make In and InOut argument by pointer
				pw_c.printf("*");
			}
			pw_c.printf(" %s", arg.getName());
		}
		pw_c.printf(")\n");
		pw_c.printf("{\n");
		pw_c.printf("    ctx->con = con_init(ctx->host, ctx->port);\n");
		pw_c.printf("    con_prep(ctx->con);\n");
		pw_c.printf("    con_put_charptr(ctx->con, \"%s\");\n", methnam);
		for(Arg arg : args.getArgs()) {
			if(arg.getType() == Arg.DataType.Block) {
				pw_h.printf("#include \"%s.h\"\n", arg.getImpl().toLowerCase());
			}
		}
		pw_h.printf("int32_t %s_%s(struct tig_t *ctx", prefix, methnam.toLowerCase());
		for(Arg arg : args.getArgs()) {
			pw_h.printf(", %s", arg.getType() != Arg.DataType.Block ? typeName(arg.getType()) : arg.getImpl().toLowerCase());
			if(arg.getUse() != Usage.In) {
				// make In and InOut argument by pointer
				pw_h.printf("*");
			}
			pw_h.printf(" %s", arg.getName());
		}
		pw_h.printf(");\n");
	}
	@Override
	public void genMethodBody(Method.ArgList args) {
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.Out || arg.getType() == DataType.Block) {
				if(arg.getType() != DataType.Block) {
			    	// for In and InOut non-blocks write value to connection
					if(arg.getType() == DataType.Word) {
						pw_c.printf("    %s = to_little_16(%s);\n", arg.getName(), arg.getName());
					}
					if(arg.getType() == DataType.LongWord) {
						pw_c.printf("    %s = to_little_32(%s);\n", arg.getName(), arg.getName());
					}
					if(arg.getType() == DataType.QuadWord) {
						pw_c.printf("    %s = to_little_64(%s);\n", arg.getName(), arg.getName());
					}
			        pw_c.printf("    con_put_%s(ctx->con, %s);\n", typeName(arg.getType()).toLowerCase(), arg.getName());
				} else {
			    	// for In and InOut blocks write value to connection
			        pw_c.printf("    con_put_%s(ctx->con, *%s, sizeof(**%s));\n", typeName(arg.getType()).toLowerCase(), arg.getName(), arg.getName());
				}
			}
		}
		pw_c.printf("    con_send(ctx->con);\n");
		pw_c.printf("    con_recv(ctx->con);\n");
		pw_c.printf("    int retval = con_get_int32_t(ctx->con);\n");
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.In) {
				if(arg.getType() != DataType.Block) {
					// for InOut and Out non-block read value from connection
					pw_c.printf("    *%s = con_get_%s(ctx->con);\n", arg.getName(), typeName(arg.getType()).toLowerCase());
					if(arg.getType() == DataType.Word) {
						pw_c.printf("    *%s = from_little_16(*%s);\n", arg.getName(), arg.getName());
					}
					if(arg.getType() == DataType.LongWord) {
						pw_c.printf("    *%s = from_little_32(*%s);\n", arg.getName(), arg.getName());
					}
					if(arg.getType() == DataType.QuadWord) {
						pw_c.printf("    *%s = from_little_64(*%s);\n", arg.getName(), arg.getName());
					}
				} else {
					// for InOut and Out non-block read value from connection
					pw_c.printf("    *%s = (%s)con_get_%s(ctx->con);\n", arg.getName(), arg.getImpl().toLowerCase(), typeName(arg.getType()).toLowerCase());
				}
			}
		}
	}
	@Override
	public void genMethodFooter(Method.ArgList args) {
		pw_c.printf("    con_close(ctx->con);\n");
		pw_c.printf("    return retval;\n");
		pw_c.printf("}\n");
		pw_c.printf("\n");
	}
	@Override
	public void genClassFooter() {
		// nothing
	}
	@Override
	public void close() {
		pw_c.close();
		pw_h.printf("\n");
		pw_h.printf("#endif\n");
		pw_h.close();
	}
}
