/*
 * Copyright 2024 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.tig.client;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import dk.vajhoej.vms.tig.Arg;
import dk.vajhoej.vms.tig.Arg.DataType;
import dk.vajhoej.vms.tig.Arg.Usage;
import dk.vajhoej.vms.tig.Method;
import dk.vajhoej.vms.tig.shared.Util;

public class JavaClientGen implements ClientGen {
	private PrintWriter pw;
	@Override
	public String fileName(String imgnam) {
		return className(imgnam) + ".java";
	}
	@Override
	public String className(String imgnam) {
		return Util.capitalize(imgnam) + "Client";
	}
	@Override
	public String typeName(Arg.DataType typ) {
		switch(typ) {
			case Byte:
				return "byte";
			case Word:
				return "short";
			case LongWord:
				return "int";
			case QuadWord:
				return "long";
			case F_Float:
			case S_Float:
				return "float";
			case D_Float:
			case G_Float:
			case T_Float:
				return "double";
			case FixedCharacterString:
			case VariableCharacterString:
			case NullTermCharacterString:
				return "String";
			case Block:
				return "Block";
			default:
				throw new RuntimeException("Unsupported type: " + typ);
		}
	}
	@Override
	public void open(String fnm) throws IOException {
		pw = new PrintWriter(new FileWriter(fnm));
	}
	@Override
	public void genClassHeader(String ns, String clznam) {
		pw.printf("// GENERATED CODE - edits may be overwritten\n");
		if(ns != null) {
			pw.printf("package %s;\n", ns);
			pw.printf("\n");
		}
		pw.printf("import java.io.IOException;\n");
		pw.printf("import java.net.Socket;\n");
		pw.printf("\n");
		pw.printf("import dk.vajhoej.record.RecordException;\n");
		pw.printf("import dk.vajhoej.vms.tig.shared.Connection;\n");
		pw.printf("\n");
		pw.printf("public class %s {\n", clznam);
		pw.printf("    private String host;\n");
		pw.printf("    private int port;\n");
	}
	@Override
	public void genCtor(String clznam, int port) {
		pw.printf("    public %s(String host) throws IOException {\n", clznam);
		pw.printf("        this.host = host;\n");
		pw.printf("        this.port = %d;\n", port);
		pw.printf("    }\n");
	}
	@Override
	public void genMethodHeader(String methnam, Method.ArgList args) {
		pw.printf("    public int %s(", methnam);
		boolean first = true;
		for(Arg arg : args.getArgs()) {
			if(first) {
				first = false;
			} else {
				pw.printf(", ");
			}
			pw.printf("%s", arg.getType() != Arg.DataType.Block ? typeName(arg.getType()) : arg.getImpl());
			if(arg.getUse() != Usage.In) {
				// make In and InOut argument arrays (Java always pass by value)
				pw.printf("[]");
			}
			pw.printf(" %s", arg.getName());
		}
		pw.printf(") throws IOException, RecordException {\n");
		pw.printf("        Socket s = new Socket(host, port);\n");
		pw.printf("        Connection con = new Connection(s);\n");
		pw.printf("        con.prep();\n");
		pw.printf("        con.putString(\"%s\");\n", methnam);
	}
	@Override
	public void genMethodBody(Method.ArgList args) {
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.Out || arg.getType() == DataType.Block) {
			    if(arg.getUse() == Usage.In) {
			    	// for In write value to connection
			        pw.printf("        con.put%s(%s);\n", Util.capitalize(typeName(arg.getType())), arg.getName());
			    } else {
			    	// for InOut write value to connection
			        pw.printf("        con.put%s(%s[0]);\n", Util.capitalize(typeName(arg.getType())), arg.getName());
			    }
			}
		}
		pw.printf("        con.send();\n");
		pw.printf("        con.recv();\n");
		pw.printf("        int retval = con.getInt();\n");
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.In) {
				if(arg.getType() != DataType.Block) {
					// for InOut and Out non-block read value from connection
					pw.printf("        %s[0] = con.get%s();\n", arg.getName(), Util.capitalize(typeName(arg.getType())));
				} else {
					// for InOut and Out non-block read value from connection (class is needed to properly deserialize)
					pw.printf("        %s[0] = (%s)con.get%s(%s.class);\n", arg.getName(), arg.getImpl(), Util.capitalize(typeName(arg.getType())), arg.getImpl());
				}
			}
		}
	}
	@Override
	public void genMethodFooter(Method.ArgList args) {
		pw.printf("        con.close();\n");
		pw.printf("        return retval;\n");
		pw.printf("    }\n");
	}
	@Override
	public void genClassFooter() {
		pw.printf("}\n");
	}
	@Override
	public void close() {
		pw.close();
	}
}
